//    WeIRCd administration tool
//    Copyright (C) 2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "admintool.h"

#define MAXAMOUNT 1024

char parsebuf[MAXAMOUNT+1] = "";
int insidelinecomment = 0;
int insideblockcomment = 0;
char combinedline[MAXAMOUNT+1];

void parser_reset() {
	strcpy(parsebuf,""); /* clear our parser's buffer */
	strcpy(combinedline,""); /* clear our line combine buffer where we put together the final lines */
	fill_start(); /* clear all vars previously read */
}
void parser_extractedline(const char* line) {
	/* we got rid of comments etc. now we have the very line itself. yeah. */
	//printf("value: %s\n",line);
	char* linecopy = alloca(strlen(line)+1);
	strcpy(linecopy,line);
	if (strendswith(line,"{") == 0) {
		linecopy[strlen(linecopy)-1] = 0;
		trimright(linecopy);
		if (strlen(linecopy) <= 0) {return;}
		fill_openblock(linecopy);return;
	}
	if (strcasecmp(line,"}") == 0) {
		fill_closeblock();return;
	}
	char part0[50];
	int cpos = charpos(line,'=');
	if (cpos > 0) {
		strcopy(part0,line,49);
		if (cpos <= 49) {part0[cpos] = 0;}
		trimright(part0);
		cut(linecopy,cpos+1);trimleft(linecopy);
		fill_entervar(part0,linecopy);
	}else{
		fill_entervar("LISTVAL",line);
	}
}
void parser_parseline(const char* line) {
	/* ok, we mainly want to get rid of comments and other ugly stuff first */
	//printf("parsing %s\n",line);
	int r = 0;
	while (r < strlen(line)) {
		if (line[r] == '\n') {
			insidelinecomment = 0;
			trimleft(combinedline);
			trimright(combinedline);
			if (strlen(combinedline) > 0) {
				parser_extractedline(combinedline);
			}
			strcpy(combinedline,"");
			r++;continue;
		}
		if (insidelinecomment == 0 && insideblockcomment == 0) {
			if (line[r] == '/') {
				if (r < strlen(line)-1) {
					if (line[r+1] == '/') {
						insidelinecomment = 1;
						r++;continue;
					}
					if (line[r+1] == '*') {
						insideblockcomment = 1;
						r+=2;continue;
					}
				}
			}
			if (line[r] == ';' || line[r] == '}' || line[r] == '{') {
				if (line[r] != ';') {
					addtobuf(combinedline,line[r],MAXAMOUNT+1);
				}
				trimleft(combinedline);
				trimright(combinedline);
				if (strlen(combinedline) > 0) {
					parser_extractedline(combinedline);
				}
				strcpy(combinedline,"");
				r++;continue;
			}
			addtobuf(combinedline,line[r],MAXAMOUNT+1);
		}else{
			if (insidelinecomment == 1) {
				if (line[r] == '\n') {
					insidelinecomment = 0;
					trimleft(combinedline);
					trimright(combinedline);
					if (strlen(combinedline) > 0) {
						parser_extractedline(combinedline);
					}
					strcpy(combinedline,"");
					r++;continue;
				}
			}
			if (insideblockcomment == 1) {
				if (line[r] == '*') {
					if (r < strlen(line)-1) {
						if (line[r+1] == '/') {
							insideblockcomment = 0;
							r+=2;continue;
						}
					}
				}
			}
		}
		r++;
	}
}
int parser_feed(const char* data, int amount) {
	if (amount <= 0) {return 0;}
	int len = strlen(parsebuf);
	int fetcheddata = 0;
	while (1) {
		len = strlen(parsebuf);
		if (len < MAXAMOUNT) { /* check how many bytes we have left */
			int i = MAXAMOUNT-len; /* that's the amount we'd like to read */
			if (i > amount) {i = amount;} /* the actual data available is probably less */
			memcpy(parsebuf+len, data, i); /* now copy that data */
			fetcheddata += i; /* we really want to remember that (to return it to the caller) */
			amount -= i;
			parsebuf[i+len] = 0; /* don't forget the null terminator */
		}else{
			/* old data is useless (no line break to be found). throw it away. */
			//printf("no line break found in full buffer.\n");
			int i = amount;
			if (i > MAXAMOUNT) {i = MAXAMOUNT;}
			fetcheddata += i; /* remember the amount of data we fetched */
			amount -= i;
			memcpy(parsebuf,data,i);parsebuf[i] = 0; /* copy it and nullterminate */
		}
		//printf("fetched %d bytes; buffer fill state: %d\n",fetcheddata,strlen(parsebuf));
		/* ok, now check for new lines */
		char tempbuf[MAXAMOUNT+1];
		int linebreakpos = charpos(parsebuf,'\n');
		if (linebreakpos >= 0) {
			/* we have line breaks, that is, complete lines to proceed: */
			while (linebreakpos >= 0) { /* go through all lines */
				strcpy(tempbuf,parsebuf);
				tempbuf[linebreakpos+1] = 0; /* create a copy and cut off at line break */
				trimleft(tempbuf);
				if (strlen(tempbuf) > 0) {
					parser_parseline(tempbuf); /* parse it */
				}
				cut(parsebuf,linebreakpos+1); /* throw away parsed parts */
				linebreakpos = charpos(parsebuf, '\n'); /* search next line's end */
			}
			/* ok, now we'll loop to see if, now as we removed something out of the buffer,
			   we can still fetch some data now we couldn't earlier (because buffer was full).
			   We might even be able to parse something of this. */
		}else{
			/* no lines left to proceed. quit. */
			return fetcheddata;
		}
	}
	return fetcheddata;
}
void parser_finish() {
	trimleft(parsebuf);
	if (strlen(parsebuf) > 0) {
		parser_parseline(parsebuf);
		strcpy(parsebuf,"");
	}
}