//    WeIRCd 0.7 IRC server software - main file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "weircd.h"

char startupprefix[] = " [****]";
char startupprefixsub[] = " [    ]";
char version_copy[50];
char bigversion[] = "0.7";
char version_detail[] = "0.7.5.1";
int willquit = 0;
char md5result[100];

char datadir[500];
char moduledir[500];

void fatalerror() {
	printf("%s FATAL ERROR, terminating.\n",startupprefix);
	#ifdef WIN
	int i = getchar();
	while (i != EOF && i != '\n') {i = getchar();}
	#endif
	#ifndef WIN
	printf("\n<-------------------------------------------->\n");
	#endif
	return;
}


void rootabort() {
	printf("\n%s Aborting...\n\n",startupprefix);
	printf("------ DO >NOT< RUN WEIRCD AS ROOT USER ------\n");
	printf("        This is dangerous. Please create\n");
	printf("        a user account that has only the\n");
	printf("        limited privilegues that an ircd\n");
	printf("        needs.\n");
	printf("----------------------------------------------\n");
	fatalerror();
}
#ifndef WIN
int configown(const char* path, int uid, int gid) {
	printf("Changing ownership of '%s' to user 'weircd'...\n",path);
	errno = 0;
	int result = chown(path,uid,gid);
	if (result < 0) {
		if (errno == EPERM || errno == EACCES) {
			printf("Failed to chown() '%s': Access denied\n",path);
			return -1;
		}
		if (errno == ENOENT || errno == ENOTDIR) {
			printf("'%s' is not present. Proceeding...\n",path);
		}else{
			printf("Failed to chown() '%s': Unknown error\n",path);
			return -1;
		}
	}
	return 0;
}
#endif

#ifndef WIN
int main (int argc, char** argv) {
	strncpy(datadir,WDATADIR,499);datadir[499] = 0;
	if (strlen(datadir) > 0) {
		if (datadir[strlen(datadir)-1] == '/') {
			datadir[strlen(datadir)-1] = 0;
		}
		strncpy(moduledir,datadir,499);moduledir[499] = 0;
		stradd(moduledir,"/modules",499);
	}else{
		strcpy(moduledir,"");
		printf("FATAL ERROR: No data dir defined.\n");
		return -1;
	}
	int i = evaluateParams_linux(argc,(const char**)argv); if (i < 1) {return i;}
	signals_Setup();
#endif
#ifdef WIN
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
			LPSTR lpCmdLine, int nCmdShow) {
	strcpy(datadir,"");strcpy(moduledir,"modules");
	int i = evaluateParams(lpCmdLine);if (i < 1) {return i;}
#endif
	
	
	if (startup_WeircdInit(version_detail) < 0) {return -1;}
	
	
	
	
#ifndef ALLOWROOT
#ifdef SAFEUNIX
	if ((getuid() == 0 || geteuid() == 0) && option_allowroot != 1) {
		printf("\nRunning as root.\n");
		printf("Checking if user named 'weircd' exists...\n");
		errno = 0;
		struct passwd* p = getpwnam("weircd");
		if (p == NULL) {
			if (errno != 0) {
				printf("Failed to access /etc/passwd.\n");
				rootabort();return -1;
			}else{
				//no such user yet, create it
				printf("Attempting to create user 'weircd'\n");
				system("useradd --shell /bin/false weircd");
				p = getpwnam("weircd");
				if (p == NULL) {
					printf("User creation failed.\n");
					rootabort();return -1;
				}
				printf("User 'weircd' with uid '%d' created.\n",p->pw_uid);
			}
		}else{
			printf("User 'weircd' with uid '%d' found.\n",p->pw_uid);
		}
		int uid = p->pw_uid;
		int gid = p->pw_gid;
		if (configown(datadir, uid, gid) < 0) {
			rootabort();return -1;
		}
		if (configown(moduledir, uid, gid) < 0) {
			rootabort();return -1;
		}
		if (configown("./weircdconf.cfg", uid, gid) < 0) {
			rootabort();return -1;
		}
		char globalconfig[50];char seperator[2];
		strcpy(seperator,"/");
		if (strendswith(CONFDIR,"/") == 0) {strcpy(seperator,"");}
		snprintf(globalconfig,49,"%s%sweircdconf.cfg",CONFDIR,seperator);globalconfig[49] = 0;
		if (configown(globalconfig, uid, gid) < 0) {
			rootabort();return -1;
		}
		printf("Changing own process uid to '%d'.\n",p->pw_uid);
		seteuid(uid);setuid(uid);
		printf("Now we should have safe normal user rights! Enjoy!\n\n");
	}
#endif
#endif
	if (readconfig(0) < 0) {
		printf("%s    > Error processing config file, Cancelling.\n",startupprefixsub);
		fatalerror();
		return -1;
	}
	printf("%s Loading glines...\n",startupprefix);
	if (readglines(0) < 0) {
		printf("%s    > Error processing G-lines file, Cancelling.\n",startupprefixsub);
		fatalerror();
		return -1;
	}
	if (readglines(1) < 0) {
		printf("%s    > Error processing K-lines file, Cancelling.\n",startupprefixsub);
		fatalerror();
		return -1;
	}
#ifndef WIN
	struct sigaction signalst;
	if (sigaction(SIGPIPE, NULL, &signalst) == 0) {
		signalst.sa_handler = SIG_IGN;
		sigaction(SIGPIPE, &signalst, NULL);
	}else{
		printf("%s Failed to mask SIGPIPE signal handler\n",startupprefix);
		fatalerror();
		return -1;
	}
#endif
	printf("%s Initialising socket system...\n",startupprefix);
	int r = so_Startup();
	if (r < 0) {
		printf("%s    > Was not able to start socket system!\n",startupprefixsub);
		fatalerror();
		return -1;
	}
	printf("%s Opening listeners...\n",startupprefix);
	r = userman_RefreshState();
	if (r < 0) {
		if (r < -1) {
			printf("%s Failed to allocate memory for opening listeners",startupprefixsub);
			fatalerror();
			return -1;
		}
		printf("%s  WARNING: Failed to open one or more ports - Continuing..\n",startupprefixsub);
	}
	//weircdLinksInit();
#ifndef WIN
	printf("%s SUCCESS, now up and running!\n",startupprefix);
	printf("\n<-------------------------------------------->\n\n");
#endif
#ifndef NOFORK
#ifndef WIN
	if (option_nofork == 0) {
		int forking = startup_ForkIntoBackground();
		if (forking < 0) {
			if (forking < -1) {
				printf("ERROR: Forking failed.\n");
				fatalerror();
				return -1;
			}else{
				return 0;
			}
		}
	}
#endif
#endif
#ifdef WIN
	printf("%s Running!\n",startupprefix);
	printf("\n(Please leave this window open)\n");
#endif
	fflush(stdout);
	return mainloop_go();
}
